;;;
;;; s c m p . s c m		Simple C Meta-Programming
;;;				with low-level hierarchical blocks of C code
;;;

;;; Simple block manipulation
(define (make-block . stuff) (cons 'R stuff))
(define (push-block! block stuff)
  (set-cdr! block (cons stuff (cdr block))))
(define (pop-block! block)
  (let ((stuff (cadr block)))
    (set-cdr! block (cddr block))
    stuff))
(define (append-block! block list)
  (if (null? (cdr block)) (set-cdr! block list)
      (append-block! (cdr block) list)))
(define (queue-block! block stuff)
  (append-block! block (list stuff)))
(define (unqueue-block! block)
  (if (null? (cddr block))
      (let ((stuff (cadr block)))
	(set-cdr! block '())
	stuff)
      (unqueue-block! (cdr block))))

;;; CPP stuff
(define (CPP:include-sys file)
  `(R "#include <" ,file ">\n"))
(define (CPP:include file)
  `(R "#include \"" ,file "\"\n"))
(define (CPP:comment . stuff)
  `(R "/*" ,@stuff "*/")) ; should quote string (?)


;;; block-oriented stuff
(define (blocks decl-block init-block text-block exit-block)
   (list text-block init-block exit-block))
(define (get-decl-block blocks) (car blocks))
(define (get-init-block blocks) (cadr blocks))
(define (get-text-block blocks) (caddr blocks))
(define (get-exit-block blocks) (cadddr blocks))
(define (add-decl! blocks text)
   (enqueue-block! get-decl-block text))
(define (add-init! blocks init)
   (enqueue-block! get-init-block init))
(define (add-text! blocks text)
   (enqueue-block! get-text-block text))
(define (add-exit! blocks exit)
   (push-block! get-exit-block exit))

;;; Final output
(define (SCMP filename block)
  (define SCMP:output-port (open-output-file filename))
  (define (SCMP:display foo)
    (display foo SCMP:output-port))
  (define (SCMP:write foo)
    (SCMP:write foo SCMP:output-port))
  (define (SCMP:output-code foo)
    (cond
     ; "..." or number: insert plain source code
     ((or (string? foo)
	  (integer? foo))
      (SCMP:display foo))
     ; '... : insert quoted object
     ((and (pair? foo) (eq? (car foo) 'quote))
      (SCMP:write (cadr foo)))
     ; (R ...): recurse
     ((and (pair? foo) (eq? (car foo) 'R))
      (for-each SCMP:output-code (cdr foo)))
     ; else: error
     (else
      (display "Bad argument in make-C-code: ")
      (DBG foo))
     ))
  (SCMP:output-code block)
  (close-output-port SCMP:output-port))
